//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "FormMain.h"
#include "FormAbout.h"
#include "FormSettings.h"
#include "FormLog.h"
#include "Log.h"
#include "Mutex.h"
#include "ScopedLock.h"
#include "GridUtils.h"
#include <mmsystem.h>

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TfrmMain *frmMain;

Mutex mtxPattern;
MMRESULT timerID = NULL;
int currentPos = -1;

//---------------------------------------------------------------------------
__fastcall TfrmMain::TfrmMain(TComponent* Owner)
    : TForm(Owner), lpt(INVALID_HANDLE_VALUE)
{

}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::FormCreate(TObject *Sender)
{
    AnsiString asConfigFile = ChangeFileExt( Application->ExeName, ".ini" );
    appSettings.Read(asConfigFile);
    if (this->BorderStyle != bsSingle)
    {
        this->Width = appSettings.frmMain.iWidth;
        this->Height = appSettings.frmMain.iHeight;
    }
    this->Top = appSettings.frmMain.iPosY;
    this->Left = appSettings.frmMain.iPosX;
    if (appSettings.frmMain.bAlwaysOnTop)
        this->FormStyle = fsStayOnTop;
    else
        this->FormStyle = fsNormal;
    if (appSettings.frmMain.bWindowMaximized)
        this->WindowState = wsMaximized;
    if (appSettings.Logging.bLogToFile)
        CLog::Instance()->SetFile(ChangeFileExt(Application->ExeName, ".log").c_str());
    else
        CLog::Instance()->SetFile("");
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::FormCloseQuery(TObject *Sender, bool &CanClose)
{
    AnsiString asConfigFile = ChangeFileExt( Application->ExeName, ".ini" );
    appSettings.frmMain.bWindowMaximized = (this->WindowState == wsMaximized);
    if (!appSettings.frmMain.bWindowMaximized)
    {
        // these values are meaningless is wnd is maximized
        appSettings.frmMain.iWidth = this->Width;
        appSettings.frmMain.iHeight = this->Height;
        appSettings.frmMain.iPosY = this->Top;
        appSettings.frmMain.iPosX = this->Left;
    }
    appSettings.Write(asConfigFile);

    CanClose = true;
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::actShowAboutExecute(TObject *Sender)
{
    frmAbout->ShowModal();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::actShowSettingsExecute(TObject *Sender)
{
    frmSettings->appSettings = &appSettings;
    frmSettings->ShowModal();
    if (appSettings.frmMain.bAlwaysOnTop)
        this->FormStyle = fsStayOnTop;
    else
        this->FormStyle = fsNormal;
    if (appSettings.Logging.bLogToFile)
        CLog::Instance()->SetFile(ChangeFileExt(Application->ExeName, ".log").c_str());
    else
        CLog::Instance()->SetFile("");
    frmLog->SetBufferSize(appSettings.Logging.iMaxUiLogLines);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::FormShow(TObject *Sender)
{
    static bool once = false;
    if (!once)
    {
        once = true;
        frmLog->SetBufferSize(appSettings.Logging.iMaxUiLogLines);
        CLog::Instance()->SetLevel(E_LOG_TRACE);
        CLog::Instance()->callbackLog = frmLog->OnLog;

		Grid->ColWidths[0] = 40;

		timeBeginPeriod(1);

		int iRowCount = StrToIntDef(edRowsCount->Text, 20);
		edRowsCount->Text = iRowCount;
		SetRowCount(iRowCount);

		tmrStartup->Enabled = true;
    }
    LOG("Application started");
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::FormDestroy(TObject *Sender)
{
    CLog::Instance()->Destroy();
	if (lpt != INVALID_HANDLE_VALUE)
	{
		CloseHandle(lpt);
		lpt = INVALID_HANDLE_VALUE;
	}
	if (timerID != NULL)
	{
		timeKillEvent(timerID);
		timerID = NULL;
	}
	timeEndPeriod(1);		
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::actShowLogExecute(TObject *Sender)
{
    if (!frmLog->Visible)
        frmLog->Show();
    frmLog->BringToFront();
}
//---------------------------------------------------------------------------

void TfrmMain::SetRowCount(int count)
{
	Grid->RowCount = count + Grid->FixedRows;
	pattern.clear();
	// X-labels
	for (int i=Grid->FixedCols; i<Grid->ColCount; i++)
	{
		Grid->Cells[i][0] = IntToStr(Grid->ColCount - (i + Grid->FixedCols));
	}
	// Y-labels
	for (int i=Grid->FixedRows; i<Grid->RowCount; i++)
	{
		unsigned char byte = 0;
		Grid->Cells[0][i] = IntToStr(i - Grid->FixedRows);
		pattern.push_back(byte);
	}
	for (int i=Grid->FixedCols; i<Grid->ColCount; i++)
	{
		for (int j=Grid->FixedRows; j<Grid->RowCount; j++)
		{
			Grid->Cells[i][j] = "";
			SetCheckBox(*Grid, i, j, true, false);
		}
	}
}


void __fastcall TfrmMain::btnOpenClick(TObject *Sender)
{
    if (lpt != INVALID_HANDLE_VALUE)
        CloseHandle(lpt);
    lpt = CreateFile(cbLpt->Text.c_str(),
                     GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, 0);
    if (lpt == INVALID_HANDLE_VALUE)
    {
        ReportError("Could not open LPT port");
        return;
    }

    /** \note
    http://www.codeguru.com/forum/showthread.php?t=231214
    The parallelTimeouts.WriteTotalTimeoutConstant variable has a minimum value of 2000 (2 seconds) on WinXP.
    Default value is set to 300000 (5 mins).
    */
    COMMTIMEOUTS parallelTimeouts;
    parallelTimeouts.ReadIntervalTimeout = MAXDWORD;
    parallelTimeouts.ReadTotalTimeoutMultiplier = 3000;
    parallelTimeouts.ReadTotalTimeoutConstant = 3000;
    parallelTimeouts.WriteTotalTimeoutMultiplier = 3000;
    parallelTimeouts.WriteTotalTimeoutConstant = 3000;

    if (!SetCommTimeouts(lpt, &parallelTimeouts))
    {
        ReportError("SetCommTimeouts failed");
        return;
    }
    SetStatus("%s opened", cbLpt->Text.c_str());
}
//---------------------------------------------------------------------------

void TfrmMain::ReportError(AnsiString caption)
{
    LPVOID lpMsgBuf;
    DWORD dw = GetLastError();
    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER |
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        dw,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0, NULL );
    MessageBox(this->Handle, (LPCTSTR)lpMsgBuf, caption.c_str(), MB_ICONEXCLAMATION);
    LocalFree(lpMsgBuf);
}

void TfrmMain::SetStatus(char *lpData, ...)
{
    AnsiString text;
    va_list ap;
    va_start(ap, lpData);
    text.vprintf(lpData, ap);
    va_end(ap);
	StatusBar->SimpleText = text;
}

void __fastcall TfrmMain::btnCloseClick(TObject *Sender)
{
    if (lpt != INVALID_HANDLE_VALUE)
        CloseHandle(lpt);
    SetStatus("Port closed");
    lpt = INVALID_HANDLE_VALUE;
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::GridDrawCell(TObject *Sender, int ACol,
                                       int ARow, TRect &Rect, TGridDrawState State)
{
	TStringGrid* StringGrid = static_cast<TStringGrid*>(Sender);
	assert(StringGrid != NULL);

	TCanvas* SGCanvas = StringGrid->Canvas;
	SGCanvas->Font = StringGrid->Font;

	RECT RText = static_cast<RECT>(Rect);
	const AnsiString text(StringGrid->Cells[ACol][ARow]);

	const bool fixed = State.Contains(gdFixed);
	const bool focused = State.Contains(gdFocused);
	bool selected = State.Contains(gdSelected);
    if (!StringGrid->Options.Contains(goDrawFocusSelected))
    {
        selected = selected && !focused;
    }
    // if the cell is fixed (headers)
    if (fixed)
    {
		SGCanvas->Brush->Color = StringGrid->FixedColor;
		SGCanvas->Font->Color = clBtnText;
		SGCanvas->FillRect(Rect);
		Frame3D(SGCanvas, Rect, clBtnHighlight, clBtnShadow, 1);
	}
	/*
	  // if the cell is selected
	  else if (selected) {
		SGCanvas->Brush->Color = clHighlight;
		SGCanvas->Font->Color = clHighlightText;
		SGCanvas->FillRect(Rect);
	  }
	*/
	// if the cell is normal
	else
	{
		SGCanvas->Brush->Color = StringGrid->Color;
		SGCanvas->Font->Color = StringGrid->Font->Color;
		SGCanvas->FillRect(Rect);
	}
	// if the cell is focused
	if (focused)
	{
		DrawFocusRect(SGCanvas->Handle, &RText);
	}

	// if this cell contains a checkbox
	if (GetCheckBox(*StringGrid, ACol, ARow))
	{
		// set the flags for rendering checked/unchecked
		unsigned int state = DFCS_BUTTONCHECK;
		if (GetCheckState(*StringGrid, ACol, ARow))
		{
			state = state | DFCS_CHECKED;
			SetBit(ARow - StringGrid->FixedRows, StringGrid->ColCount - ACol - StringGrid->FixedCols, true);
		}
		else
		{
			SetBit(ARow - StringGrid->FixedRows, StringGrid->ColCount - ACol - StringGrid->FixedCols, false);
		}

		// size the checkbox
		RECT RCell =static_cast<RECT>(Rect);
		OffsetRect(&RCell, 2, (int)(0.5 * (RCell.bottom - RCell.top)));
		RCell.right = RCell.left + GetSystemMetrics(SM_CXMENUCHECK);
		RCell.bottom = RCell.top + GetSystemMetrics(SM_CYMENUCHECK);
		RCell.top -= (int)(0.5 * (RCell.bottom - RCell.top)) + 2;

		// draw the checkbox
		DrawFrameControl(StringGrid->Canvas->Handle, &RCell, DFC_BUTTON, state);

		// move the text over
		RText.left = RCell.right;
	}

	// draw the text
	RText.left += 2;
	RText.top += 2;
	DrawText(SGCanvas->Handle, text.c_str(), text.Length(), &RText, DT_LEFT |DT_VCENTER |DT_SINGLELINE);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::GridMouseDown(TObject *Sender, TMouseButton Button, TShiftState Shift, int X, int Y)
{
	if (Button != mbLeft)
	{
    	return;
	}

	TStringGrid* StringGrid = static_cast<TStringGrid*>(Sender);
	assert(StringGrid != NULL);

	int Col, Row;
	if (PtInCheckBox(*StringGrid, X, Y, Col, Row))
	{
		if (GetCheckBox(*StringGrid, Col, Row))
		{
			bool is_checked = GetCheckState(*StringGrid, Col, Row);
			SetCheckState(*StringGrid, Col, Row, !is_checked);
		}
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::GridKeyPress(TObject *Sender, char &Key)
{
	TStringGrid* StringGrid = static_cast<TStringGrid*>(Sender);
    assert(StringGrid != NULL);

    if (Key == VK_SPACE)
    {
        const int col = StringGrid->Col;
        const int row = StringGrid->Row;

		if (GetCheckBox(*StringGrid, col, row))
		{
			SetCheckState(*StringGrid, col, row, !GetCheckState(*StringGrid, col, row));
		}
	}
}
//---------------------------------------------------------------------------

void TfrmMain::SetBit(int pos, int bit, bool state)
{
	ScopedLock<Mutex> lock(mtxPattern);
	if (state)
	{
		pattern[pos] |= (unsigned char)1 << bit;
	}
	else
	{
		pattern[pos] &= ~(unsigned char)1 << bit;
	}
}

void CALLBACK TimeProc(UINT uID, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2)
{
   	ScopedLock<Mutex> lock(mtxPattern);
	if (frmMain->lpt == INVALID_HANDLE_VALUE)
		return;
	currentPos++;
	if (currentPos >= frmMain->pattern.size())
	{
    	currentPos = 0;
	}
	unsigned char byte = frmMain->pattern[currentPos];
	DWORD dwBytesWritten;
	if (!WriteFile(frmMain->lpt, &byte, 1, &dwBytesWritten, NULL))
	{
		frmMain->ReportError("Write error");
		if (timerID != NULL)
		{
			timeKillEvent(timerID);
			timerID = NULL;
		}
	}
}

void __fastcall TfrmMain::btnStartClick(TObject *Sender)
{
	currentPos = -1;
	if (timerID != NULL)
	{
		timeKillEvent(timerID);
	}
	int newPeriod = StrToIntDef(edStepPeriod->Text, 500);
	edStepPeriod->Text = newPeriod;
	timerID = timeSetEvent(newPeriod, 0, TimeProc, 0, TIME_PERIODIC | TIME_KILL_SYNCHRONOUS);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnStopClick(TObject *Sender)
{
	if (timerID != NULL)
	{
		timeKillEvent(timerID);
		timerID = NULL;
	}
	currentPos = -1;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnSetRowsCountClick(TObject *Sender)
{
	int iRowCount = StrToIntDef(edRowsCount->Text, 20);
	edRowsCount->Text = iRowCount;
	SetRowCount(iRowCount);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::tmrStartupTimer(TObject *Sender)
{
	tmrStartup->Enabled = false;
	btnOpenClick(NULL);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::tmrUpdatePositionTimer(TObject *Sender)
{
	frmMain->lblCurrentPositionVal->Caption = currentPos;	
}
//---------------------------------------------------------------------------

